@@ -1,12 +1,12 @@
-/*  Copyright (C) 2014-2017 CZ.NIC, z.s.p.o. <knot-dns@labs.nic.cz>
+/*  Copyright (C) 2014-2021 CZ.NIC, z.s.p.o. <knot-dns@labs.nic.cz>
  *  SPDX-License-Identifier: GPL-3.0-or-later
  */
 
 #pragma once
 
-#include <assert.h>
 #include <dirent.h>
 #include <stdio.h>
+#include <stdlib.h>
 #include <stdbool.h>
 #include <sys/socket.h>
 #include <sys/time.h>
@@ -26,9 +26,8 @@
 #include "contrib/mempattern.h"
 #include "lib/defines.h"
 #include "lib/generic/array.h"
+#include "lib/log.h"
 
-struct kr_query;
-struct kr_request;
 
 /*
  * Logging and debugging.
@@ -41,60 +40,59 @@ typedef void (*trace_callback_f)(struct kr_request *request);
  * @param[in] msg Log message. Pointer is not valid after handler returns. */
 typedef void (*trace_log_f)(const struct kr_request *request, const char *msg);
 
-#define kr_log_info printf
-#define kr_log_error(...) fprintf(stderr, ## __VA_ARGS__)
-#define kr_log_deprecate(...) fprintf(stderr, "deprecation WARNING: " __VA_ARGS__)
+/** Assert() but always, regardless of -DNDEBUG.  See also kr_assert(). */
+#define kr_require(expression) do { if (!(expression)) { \
+		kr_fail(true, #expression, __func__, __FILE__, __LINE__); \
+		__builtin_unreachable(); /* aid code analysis */ \
+	} } while (false)
 
-/* Always export these, but override direct calls by macros conditionally. */
-/** Whether in --verbose mode.  Only use this for reading. */
-KR_EXPORT extern bool kr_verbose_status;
-
-/** Set --verbose mode.  Not available if compiled with -DNOVERBOSELOG. */
-KR_EXPORT bool kr_verbose_set(bool status);
-
-/**
- * @brief Return true if the query has request log handler installed.
+/** Check an assertion that's recoverable. Return the true if it fails and needs handling.
+ *
+ * If the check fails, optionally fork()+abort() to generate coredump
+ * and continue running in parent process.  Return value must be handled to
+ * ensure safe recovery from error.  Use kr_require() for unrecoverable checks.
+ * The errno variable is not mangled, e.g. you can: if (kr_fails_assert(...)) return errno;
  */
-#define kr_log_rtrace_enabled(req) (__builtin_expect( \
-	(req) && (req)->trace_log, \
-	false))
+#define kr_fails_assert(expression) !kr_assert_func((expression), #expression, \
+						   __func__, __FILE__, __LINE__)
 
-#define kr_log_qtrace_enabled(qry) (__builtin_expect( \
-	(qry) && kr_log_rtrace_enabled(qry->request), \
-	false))
+/** Kresd assertion without a return value.
+ *
+ * These can be turned on or off, for mandatory unrecoverable checks, use kr_require().
+ * For recoverable checks, use kr_fails_assert().
+ * */
+#define kr_assert(expression) (void)!kr_fails_assert((expression))
 
-/**
- * Log a message through the request log handler or stdout.
- * Caller is responsible for detecting verbose mode, use QRVERBOSE() macro.
- * @param  qry_uid query ID to append to request ID, 0 means "no query"
- * @param  indent level of indentation between [req.qry][source] and message
- * @param  source message source
- * @param  fmt message format
+/** Whether kr_assert() and kr_fails_assert() checks should abort. */
+KR_EXPORT extern bool kr_dbg_assertion_abort;
+
+/** How often kr_assert() should fork the process before issuing abort (if configured).
+ *
+ * This can be useful for debugging rare edge-cases in production.
+ * if (kr_debug_assertion_abort && kr_debug_assertion_fork), it is
+ * possible to both obtain a coredump (from forked child) and recover from the
+ * non-fatal error in the parent process.
+ *
+ * == 0 (false): no forking
+ * > 0: minimum delay between forks
+ *      (in milliseconds, each instance separately, randomized +-25%)
+ * < 0: no rate-limiting (not recommended)
  */
-KR_EXPORT KR_PRINTF(5)
-void kr_log_req(const struct kr_request * const req, uint32_t qry_uid,
-		const unsigned int indent, const char *source, const char *fmt, ...);
+KR_EXPORT extern int kr_dbg_assertion_fork;
 
-/**
- * Log a message through the request log handler or stdout.
- * Caller is responsible for detecting verbose mode, use QRVERBOSE() macro.
- * @param  qry current query
- * @param  source message source
- * @param  fmt message format
- */
-KR_EXPORT KR_PRINTF(3)
-void kr_log_q(const struct kr_query *qry, const char *source, const char *fmt, ...);
+/** Use kr_require(), kr_assert() or kr_fails_assert() instead of directly this function. */
+KR_EXPORT KR_COLD void kr_fail(bool is_fatal, const char* expr, const char *func,
+				const char *file, int line);
 
-#ifdef NOVERBOSELOG
-/* Efficient compile-time disabling of verbose messages. */
-#define kr_verbose_status false
-#define kr_verbose_set(x)
-#endif
-
-/** Block run in --verbose mode; optimized when not run. */
-#define VERBOSE_STATUS __builtin_expect(kr_verbose_status, false)
-#define WITH_VERBOSE(query) if(__builtin_expect(kr_verbose_status || kr_log_qtrace_enabled(query), false))
-#define kr_log_verbose if(VERBOSE_STATUS) printf
+/** Use kr_require(), kr_assert() or kr_fails_assert() instead of directly this function. */
+__attribute__ ((warn_unused_result))
+static inline bool kr_assert_func(bool result, const char *expr, const char *func,
+				  const char *file, int line)
+{
+	if (!result)
+		kr_fail(false, expr, func, file, line);
+	return result;
+}
 
 #define KR_DNAME_GET_STR(dname_str, dname) \
 	char dname_str[KR_DNAME_STR_MAXLEN]; \
@@ -106,13 +104,8 @@ void kr_log_q(const struct kr_query *qry, const char *source, const char *fmt, .
 	knot_rrtype_to_string((rrtype), rrtype_str, sizeof(rrtype_str)); \
 	rrtype_str[sizeof(rrtype_str) - 1] = 0;
 
-/* C11 compatibility, but without any implementation so far. */
-#ifndef static_assert
-#define static_assert(cond, msg)
-#endif
-
-// Use this for alocations with mm.
-// Use mm_alloc for alocations into mempool
+// Use this for allocations with mm.
+// Use mm_alloc for allocations into mempool
 
 /** A strcmp() variant directly usable for qsort() on an array of strings. */
 static inline int strcmp_p(const void *p1, const void *p2)
@@ -120,21 +113,6 @@ static inline int strcmp_p(const void *p1, const void *p2)
 	return strcmp(*(char * const *)p1, *(char * const *)p2);
 }
 
-
-/** Return time difference in miliseconds.
-  * @note based on the _BSD_SOURCE timersub() macro */
-static inline long time_diff(struct timeval *begin, struct timeval *end) {
-    struct timeval res = {
-        .tv_sec = end->tv_sec - begin->tv_sec,
-        .tv_usec = end->tv_usec - begin->tv_usec
-    };
-    if (res.tv_usec < 0) {
-        --res.tv_sec;
-        res.tv_usec += 1000000;
-    }
-    return res.tv_sec * 1000 + res.tv_usec / 1000;
-}
-
 /** Get current working directory with fallback value. */
 static inline void get_workdir(char *out, size_t len) {
 	if(getcwd(out, len) == NULL) {
@@ -170,6 +148,14 @@ typedef struct ranked_rr_array_entry ranked_rr_array_entry_t;
 typedef array_t(ranked_rr_array_entry_t *) ranked_rr_array_t;
 /* @endcond */
 
+typedef struct kr_http_header_array_entry {
+	char* name;
+	char* value;
+} kr_http_header_array_entry_t;
+
+/** Array of HTTP headers for DoH. */
+typedef array_t(kr_http_header_array_entry_t) kr_http_header_array_t;
+
 /** Concatenate N strings. */
 KR_EXPORT
 char* kr_strcatdup(unsigned n, ...);
@@ -189,7 +175,7 @@ static inline uint64_t kr_rand_bytes(unsigned int size)
 {
 	uint64_t result;
 	if (size <= 0 || size > sizeof(result)) {
-		kr_log_error("kr_rand_bytes(): EINVAL\n");
+		kr_log_error(SYSTEM, "kr_rand_bytes(): EINVAL\n");
 		abort();
 	}
 	uint8_t data[sizeof(result)];
@@ -247,13 +233,19 @@ int kr_pkt_put(knot_pkt_t *pkt, const knot_dname_t *name, uint32_t ttl,
 KR_EXPORT
 void kr_pkt_make_auth_header(knot_pkt_t *pkt);
 
-/** Simple storage for IPx address or AF_UNSPEC. */
-union inaddr {
+/** Simple storage for IPx address and their ports or AF_UNSPEC. */
+union kr_sockaddr {
 	struct sockaddr ip;
 	struct sockaddr_in ip4;
 	struct sockaddr_in6 ip6;
 };
 
+/** Simple storage for IPx addresses. */
+union kr_in_addr {
+	struct in_addr ip4;
+	struct in6_addr ip6;
+};
+
 /** Address bytes for given family. */
 KR_EXPORT KR_PURE
 const char *kr_inaddr(const struct sockaddr *addr);
@@ -296,8 +288,8 @@ int kr_ntop_str(int family, const void *src, uint16_t port, char *buf, size_t *b
  */
 static inline char *kr_straddr(const struct sockaddr *addr)
 {
-	assert(addr != NULL);
-	/* We are the sinle-threaded application */
+	if (kr_fails_assert(addr)) return NULL;
+	/* We are the single-threaded application */
 	static char str[INET6_ADDRSTRLEN + 1 + 5 + 1];
 	size_t len = sizeof(str);
 	int ret = kr_inaddr_str(addr, str, &len);
@@ -351,6 +343,12 @@ int kr_straddr_join(const char *addr, uint16_t port, char *buf, size_t *buflen);
 KR_EXPORT KR_PURE
 int kr_bitcmp(const char *a, const char *b, int bits);
 
+/** Masks bits. The specified number of bits in `a` from the left (network order)
+ * will remain their original value, while the rest will be set to zero.
+ * This is useful for storing network addresses in a trie. */
+KR_EXPORT
+void kr_bitmask(unsigned char *a, size_t a_len, int bits);
+
 /** @internal RR map flags. */
 static const uint8_t KEY_FLAG_RRSIG = 0x02;
 static inline uint8_t KEY_FLAG_RANK(const char *key)
@@ -400,6 +398,11 @@ int kr_ranked_rrarray_set_wire(ranked_rr_array_t *array, bool to_wire,
 			       uint32_t qry_uid, bool check_dups,
 			       bool (*extraCheck)(const ranked_rr_array_entry_t *));
 
+
+/** Style used by the kr_*_text() functions. */
+KR_EXPORT extern
+const knot_dump_style_t KR_DUMP_STYLE_DEFAULT;
+
 /**
  * @return Newly allocated string representation of packet.
  * Caller has to free() returned string.
@@ -439,7 +442,7 @@ char *kr_module_call(struct kr_context *ctx, const char *module, const char *pro
 /** Return the (covered) type of an nonempty RRset. */
 static inline uint16_t kr_rrset_type_maysig(const knot_rrset_t *rr)
 {
-	assert(rr && rr->rrs.count && rr->rrs.rdata);
+	kr_require(rr && rr->rrs.count && rr->rrs.rdata);
 	uint16_t type = rr->type;
 	if (type == KNOT_RRTYPE_RRSIG)
 		type = knot_rrsig_type_covered(rr->rrs.rdata);
@@ -477,7 +480,8 @@ static inline int kr_dname_lf(uint8_t *dst, const knot_dname_t *src, bool add_wi
 		return kr_error(EINVAL);
 	}
 	int len = right_aligned_dname_start[0];
-	assert(right_aligned_dname_start + 1 + len - KNOT_DNAME_MAXLEN == right_aligned_dst);
+	if (kr_fails_assert(right_aligned_dname_start + 1 + len - KNOT_DNAME_MAXLEN == right_aligned_dst))
+		return kr_error(EINVAL);
 	memcpy(dst + 1, right_aligned_dname_start + 1, len);
 	if (add_wildcard) {
 		if (len + 2 > KNOT_DNAME_MAXLEN)
@@ -490,6 +494,43 @@ static inline int kr_dname_lf(uint8_t *dst, const knot_dname_t *src, bool add_wi
 	return KNOT_EOK;
 }
 
+
+/** Timer, i.e stop-watch. */
+typedef struct timespec kr_timer_t;
+
+/** Start, i.e. set the reference point. */
+static inline void kr_timer_start(kr_timer_t *start)
+{
+	/* The call should be very reliable, but let's check it in _start() at least. */
+	kr_require(start && clock_gettime(CLOCK_MONOTONIC, start) == 0);
+}
+
+/** Get elapsed time in floating-point seconds. */
+static inline double kr_timer_elapsed(kr_timer_t *start)
+{
+	kr_require(start);
+	kr_timer_t end = { 0 };
+	(void)clock_gettime(CLOCK_MONOTONIC, &end);
+	return (end.tv_sec - start->tv_sec) + (double)(end.tv_nsec - start->tv_nsec) / 1e9;
+}
+
+/** Get elapsed time in micro-seconds. */
+static inline uint64_t kr_timer_elapsed_us(kr_timer_t *start)
+{
+	kr_require(start);
+	kr_timer_t end = { 0 };
+	(void)clock_gettime(CLOCK_MONOTONIC, &end);
+	// avoid negative differences, because of integer division
+	if (end.tv_nsec - start->tv_nsec < 0) {
+		end.tv_nsec += 1000*1000*1000;
+		end.tv_sec  -= 1;
+	}
+	return (uint64_t)(end.tv_sec - start->tv_sec) * 1000000
+		// adding 500 gives us rounding
+		+ (end.tv_nsec - start->tv_nsec + 500) / 1000;
+}
+
+
 /**
  * Difference between two calendar times specified as strings.
  * \param[in]  format format for strptime
