@@ -2,6 +2,9 @@
 local kres = require('kres')
 local ffi = require('ffi')
 
+local LOG_GRP_POLICY_TAG = ffi.string(ffi.C.kr_log_grp2name(ffi.C.LOG_GRP_POLICY))
+local LOG_GRP_REQDBG_TAG = ffi.string(ffi.C.kr_log_grp2name(ffi.C.LOG_GRP_REQDBG))
+
 local todname = kres.str2dname -- not available during module load otherwise
 
 -- Counter of unique rules
@@ -54,6 +57,17 @@ local function addr2sock(target, default_port)
 	return sock
 end
 
+-- Debug logging for taken policy actions
+local function log_policy_action(req, name)
+	if ffi.C.kr_log_is_debug_fun(ffi.C.LOG_GRP_POLICY, req) then
+		local qry = req:current()
+		ffi.C.kr_log_req1(
+			req, qry.uid, 2, ffi.C.LOG_GRP_POLICY, LOG_GRP_POLICY_TAG,
+			"%s applied for %s %s\n",
+			name, kres.dname2str(qry.sname), kres.tostring.type[qry.stype])
+	end
+end
+
 -- policy functions are defined below
 local policy = {}
 
@@ -77,18 +91,13 @@ function policy.MIRROR(target)
 end
 
 -- Override the list of nameservers (forwarders)
-local function set_nslist(qry, list)
+local function set_nslist(req, list)
 	local ns_i = 0
 	for _, ns in ipairs(list) do
-		-- kr_nsrep_set() can return kr_error(ENOENT), it's OK
-		if ffi.C.kr_nsrep_set(qry, ns_i, ns) == 0 then
+		if ffi.C.kr_forward_add_target(req, ns) == 0 then
 			ns_i = ns_i + 1
 		end
 	end
-	-- If less than maximum NSs, insert guard to terminate the list
-	if ns_i < 3 then
-		assert(ffi.C.kr_nsrep_set(qry, ns_i, nil) == 0);
-	end
 	if ns_i == 0 then
 		-- would use assert() but don't want to compose the message if not triggered
 		error('no usable address in NS set (check net.ipv4 and '
@@ -102,7 +111,6 @@ function policy.STUB(target)
 	if type(target) == 'table' then
 		for _, v in pairs(target) do
 			table.insert(list, addr2sock(v, 53))
-			assert(#list <= 4, 'at most 4 STUB targets are supported')
 		end
 	else
 		table.insert(list, addr2sock(target, 53))
@@ -112,7 +120,7 @@ function policy.STUB(target)
 		-- Switch mode to stub resolver, do not track origin zone cut since it's not real authority NS
 		qry.flags.STUB = true
 		qry.flags.ALWAYS_CUT = false
-		set_nslist(qry, list)
+		set_nslist(req, list)
 		return state
 	end
 end
@@ -123,7 +131,6 @@ function policy.FORWARD(target)
 	if type(target) == 'table' then
 		for _, v in pairs(target) do
 			table.insert(list, addr2sock(v, 53))
-			assert(#list <= 4, 'at most 4 FORWARD targets are supported')
 		end
 	else
 		table.insert(list, addr2sock(target, 53))
@@ -136,7 +143,7 @@ function policy.FORWARD(target)
 		qry.flags.ALWAYS_CUT = false
 		qry.flags.NO_MINIMIZE = true
 		qry.flags.AWAIT_CUT = true
-		set_nslist(qry, list)
+		set_nslist(req, list)
 		return state
 	end
 end
@@ -145,8 +152,6 @@ end
 function policy.TLS_FORWARD(targets)
 	if type(targets) ~= 'table' or #targets < 1 then
 		error('TLS_FORWARD argument must be a non-empty table')
-	elseif #targets > 4 then
-		error('TLS_FORWARD supports at most four targets (in a single call)')
 	end
 
 	local sockaddr_c_set = {}
@@ -182,7 +187,7 @@ function policy.TLS_FORWARD(targets)
 		qry.flags.AWAIT_CUT = true
 		req.options.TCP = true
 		qry.flags.TCP = true
-		set_nslist(qry, nslist)
+		set_nslist(req, nslist)
 		return state
 	end
 end
@@ -193,7 +198,13 @@ function policy.REROUTE(tbl, names)
 	local ren = require('kres_modules.renumber')
 	local prefixes = {}
 	for from, to in pairs(tbl) do
-		table.insert(prefixes, names and ren.name(from, to) or ren.prefix(from, to))
+		local prefix = names and ren.name(from, to) or ren.prefix(from, to)
+		local bitlen = prefix[2]
+		if bitlen ~= nil and bitlen % 8 ~= 0 then
+			log_warn(ffi.C.LOG_GRP_POLICY,
+				'REROUTE: network mask - only /8, /16, /24 etc. are supported (entire octets are rewritten)')
+		end
+		table.insert(prefixes, prefix)
 	end
 	-- Return rule closure
 	return ren.rule(prefixes)
@@ -202,54 +213,67 @@ end
 -- Set and clear some query flags
 function policy.FLAGS(opts_set, opts_clear)
 	return function(_, req)
+		-- We assume to be running in the begin phase, so to truly apply
+		-- to the whole request we need to change both kr_request and kr_query.
 		local qry = req:current()
-		ffi.C.kr_qflags_set  (qry.flags, kres.mk_qflags(opts_set   or {}))
-		ffi.C.kr_qflags_clear(qry.flags, kres.mk_qflags(opts_clear or {}))
+		for _, flags in pairs({qry.flags, req.options}) do
+			ffi.C.kr_qflags_set  (flags, kres.mk_qflags(opts_set   or {}))
+			ffi.C.kr_qflags_clear(flags, kres.mk_qflags(opts_clear or {}))
+		end
 		return nil -- chain rule
 	end
 end
 
+local function mkauth_soa(answer, dname, mname, ttl)
+	if mname == nil then
+		mname = dname
+	end
+	return answer:put(dname, ttl or 10800, answer:qclass(), kres.type.SOA,
+		mname .. '\6nobody\7invalid\0\0\0\0\1\0\0\14\16\0\0\4\176\0\9\58\128\0\0\42\48')
+end
+
 -- Create answer with passed arguments
 function policy.ANSWER(rtable, nodata)
 	return function(_, req)
 		local qry = req:current()
+		local data = rtable[qry.stype]
+		if data == nil and nodata ~= true then
+			return nil
+		end
+		-- now we're certain we want to generate an answer
+
 		local answer = req:ensure_answer()
 		if answer == nil then return nil end
-		local data = rtable[qry.stype]
-
 		ffi.C.kr_pkt_make_auth_header(answer)
+		local ttl = (data or {}).ttl or 1
+		answer:rcode(kres.rcode.NOERROR)
+		req:set_extended_error(kres.extended_error.FORGED, "5DO5")
 
-		if data == nil then
-			if nodata == true then
-				answer:rcode(kres.rcode.NOERROR)
-				return kres.DONE
+		if data == nil then -- want NODATA, i.e. just a SOA
+			answer:begin(kres.section.AUTHORITY)
+			local soa = rtable[kres.type.SOA]
+			if soa ~= nil then
+				answer:put(qry.sname, soa.ttl or ttl, qry.sclass, kres.type.SOA,
+							soa.rdata[1] or soa.rdata)
+			else
+				mkauth_soa(answer, kres.dname2wire(qry.sname), nil, ttl)
 			end
+			log_policy_action(req, 'ANSWER (nodata)')
 		else
-			local ttl = data.ttl or 1
-
-			answer:rcode(kres.rcode.NOERROR)
 			answer:begin(kres.section.ANSWER)
 			if type(data.rdata) == 'table' then
-				for _, rdato in ipairs(data.rdata) do
-					answer:put(qry.sname, ttl, qry.sclass, qry.stype, rdato)
+				for _, entry in ipairs(data.rdata) do
+					answer:put(qry.sname, ttl, qry.sclass, qry.stype, entry)
 				end
 			else
 				answer:put(qry.sname, ttl, qry.sclass, qry.stype, data.rdata)
 			end
-
-			return kres.DONE
+			log_policy_action(req, 'ANSWER (forged)')
 		end
+		return kres.DONE
 	end
 end
 
-local function mkauth_soa(answer, dname, mname)
-	if mname == nil then
-		mname = dname
-	end
-	return answer:put(dname, 10800, answer:qclass(), kres.type.SOA,
-		mname .. '\6nobody\7invalid\0\0\0\0\1\0\0\14\16\0\0\4\176\0\9\58\128\0\0\42\48')
-end
-
 local dname_localhost = todname('localhost.')
 
 -- Rule for localhost. zone; see RFC6303, sec. 3
@@ -340,7 +364,20 @@ function policy.all(action)
 	return function(_, _) return action end
 end
 
--- Requests which QNAME matches given zone list (i.e. suffix match)
+-- Requests whose QNAME is exactly the provided domain
+function policy.domains(action, dname_list)
+	return function(_, query)
+		local qname = query:name()
+		for _, dname in ipairs(dname_list) do
+			if ffi.C.knot_dname_is_equal(qname, dname) then
+				return action
+			end
+		end
+		return nil
+	end
+end
+
+-- Requests whose QNAME matches given zone list (i.e. suffix match)
 function policy.suffix(action, zone_list)
 	local AC = require('ahocorasick')
 	local tree = AC.create(zone_list)
@@ -400,13 +437,18 @@ local function rpz_parse(action, path)
 	local rrtype_bad = {
 		[kres.type.DNAME]  = true,
 		[kres.type.NS]     = false,
-		[kres.type.SOA]    = false,
 		[kres.type.DNSKEY] = true,
 		[kres.type.DS]     = true,
 		[kres.type.RRSIG]  = true,
 		[kres.type.NSEC]   = true,
 		[kres.type.NSEC3]  = true,
 	}
+
+	-- We generally don't know what zone should be in the file; we try to detect it.
+	-- Fortunately, it's typical that SOA is the first record, even required for AXFR.
+	local origin_soa = nil
+	local warned_soa, warned_bailiwick
+
 	local parser = require('zonefile').new()
 	local ok, errstr = parser:open(path)
 	if not ok then
@@ -415,7 +457,8 @@ local function rpz_parse(action, path)
 	while true do
 		ok, errstr = parser:parse()
 		if errstr then
-			log('[poli] RPZ %s:%d: %s', path, tonumber(parser.line_counter), errstr)
+			log_warn(ffi.C.LOG_GRP_POLICY, 'RPZ %s:%d: %s',
+				path, tonumber(parser.line_counter), errstr)
 		end
 		if not ok then break end
 
@@ -423,41 +466,62 @@ local function rpz_parse(action, path)
 		local rdata = ffi.string(parser.r_data, parser.r_data_length)
 		ffi.C.knot_dname_to_lower(full_name)
 
-		local prefix_labels = ffi.C.knot_dname_in_bailiwick(full_name, parser.zone_origin)
+		local origin = origin_soa or parser.zone_origin
+		local prefix_labels = ffi.C.knot_dname_in_bailiwick(full_name, origin)
 		if prefix_labels < 0 then
-			log('[poli] RPZ %s:%d: RR owner "%s" outside the zone (ignored)',
-				path, tonumber(parser.line_counter), kres.dname2str(full_name))
+			if not warned_bailiwick then
+				warned_bailiwick = true
+				log_warn(ffi.C.LOG_GRP_POLICY,
+					'RPZ %s:%d: RR owner "%s" outside the zone (ignored; reported once per file)',
+					path, tonumber(parser.line_counter), kres.dname2str(full_name))
+			end
 			goto continue
 		end
 
-		local bytes = ffi.C.knot_dname_size(full_name) - ffi.C.knot_dname_size(parser.zone_origin)
+		local bytes = ffi.C.knot_dname_size(full_name) - ffi.C.knot_dname_size(origin)
 		local name = ffi.string(full_name, bytes) .. '\0'
 
 		if parser.r_type == kres.type.CNAME then
 			if action_map[rdata] then
 				rules[name] = action_map[rdata]
 			else
-				log('[poli] RPZ %s:%d: CNAME with custom target in RPZ is not supported yet (ignored)',
+				log_warn(ffi.C.LOG_GRP_POLICY,
+					'RPZ %s:%d: CNAME with custom target in RPZ is not supported yet (ignored)',
 					path, tonumber(parser.line_counter))
 			end
 		else
 			if #name then
 				local is_bad = rrtype_bad[parser.r_type]
+
+				if parser.r_type == kres.type.SOA then
+					if origin_soa == nil then
+						origin_soa = ffi.gc(ffi.C.knot_dname_copy(parser.r_owner, nil), ffi.C.free)
+						goto continue -- we don't want to modify `new_actions`
+					else
+						is_bad = true -- maybe provide more info, but it seems rare
+					end
+				elseif origin_soa == nil and not warned_soa then
+					warned_soa = true
+					log_warn(ffi.C.LOG_GRP_POLICY,
+						'RPZ %s:%d warning: SOA missing as the first record',
+						path, tonumber(parser.line_counter))
+				end
+
 				if is_bad == true or (is_bad == false and prefix_labels ~= 0) then
-					log('[poli] RPZ %s:%d warning: RR type %s is not allowed in RPZ (ignored)',
+					log_warn(ffi.C.LOG_GRP_POLICY, 'RPZ %s:%d warning: RR type %s is not allowed in RPZ (ignored)',
 						path, tonumber(parser.line_counter), kres.tostring.type[parser.r_type])
 				elseif is_bad == nil then
 					if new_actions[name] == nil then new_actions[name] = {} end
 					local act = new_actions[name][parser.r_type]
 					if act == nil then
 						new_actions[name][parser.r_type] = { ttl=parser.r_ttl, rdata=rdata }
-					else -- mutiple RRs: no reordering or deduplication
+					else -- multiple RRs: no reordering or deduplication
 						if type(act.rdata) ~= 'table' then
 							act.rdata = { act.rdata }
 						end
 						table.insert(act.rdata, rdata)
 						if parser.r_ttl ~= act.ttl then -- be conservative
-							log('[poli] RPZ %s:%d warning: different TTLs in a set (minimum taken)',
+							log_warn(ffi.C.LOG_GRP_POLICY, 'RPZ %s:%d warning: different TTLs in a set (minimum taken)',
 								path, tonumber(parser.line_counter))
 							act.ttl = math.min(act.ttl, parser.r_ttl)
 						end
@@ -511,17 +575,15 @@ function policy.rpz(action, path, watch)
 					-- Watcher will also fire for changes to the directory itself
 					if name == file then
 						-- If the file changes then reparse and replace the existing ruleset
-						if verbose() then
-							log('[poli] RPZ reloading: ' .. name)
-						end
+						log_info(ffi.C.LOG_GRP_POLICY, 'RPZ reloading: ' .. name)
 						rules = rpz_parse(action, path)
 					end
 				end
 			end)
 		elseif watch then -- explicitly requested and failed
 			error('[poli] lua-cqueues required to watch and reload RPZ file')
-		elseif verbose() then
-			log('[poli] lua-cqueues required to watch and reload RPZ file, continuing without watching')
+		else
+			log_info(ffi.C.LOG_GRP_POLICY, 'lua-cqueues required to watch and reload RPZ file, continuing without watching')
 		end
 	end
 
@@ -557,7 +619,7 @@ function policy.slice_randomize_psl(seed)
 	end
 	-- load psl
 	local has_latest, psl = pcall(psl_lib.latest)
-	if not has_latest then -- compatiblity with lua-psl < 0.15
+	if not has_latest then -- compatibility with lua-psl < 0.15
 		psl = psl_lib.builtin()
 	end
 
@@ -592,7 +654,7 @@ function policy.slice_randomize_psl(seed)
 			rand_seed = rand_seed + reg_domain:byte(i)
 		end
 
-		-- use lineral congruential generator with values from ANSI C
+		-- use linear congruential generator with values from ANSI C
 		rand_seed = rand_seed % 0x80000000  -- ensure seed is positive 32b int
 		local rand = (1103515245 * rand_seed + 12345) % 0x10000
 		return 1 + rand % length
@@ -612,13 +674,18 @@ local function answer_clear(req)
 	local pkt = req:ensure_answer()
 	if pkt == nil then return nil end
 	pkt:clear_payload()
+	req:ensure_edns()
 	return pkt
 end
 
-function policy.DENY_MSG(msg)
+function policy.DENY_MSG(msg, extended_error)
 	if msg and (type(msg) ~= 'string' or #msg >= 255) then
 		error('DENY_MSG: optional msg must be string shorter than 256 characters')
         end
+	if extended_error == nil then
+		extended_error = kres.extended_error.BLOCKED
+	end
+	local action_name = msg and 'DENY_MSG' or 'DENY'
 
 	return function (_, req)
 		-- Write authority information
@@ -634,6 +701,8 @@ function policy.DENY_MSG(msg)
 				   string.char(#msg) .. msg)
 
 		end
+		req:set_extended_error(extended_error, "CR36")
+		log_policy_action(req, action_name)
 		return kres.DONE
 	end
 end
@@ -644,31 +713,57 @@ end
 
 local debug_logline_cb = ffi.cast('trace_log_f', function (_, msg)
 	jit.off(true, true) -- JIT for (C -> lua)^2 nesting isn't allowed
-	-- msg typically ends with newline
-	io.write(ffi.string(msg))
+	ffi.C.kr_log_fmt(
+		ffi.C.LOG_GRP_REQDBG, -- but the original [group] tag also remains in the string
+		LOG_DEBUG,
+		'CODE_FILE=policy.lua', 'CODE_LINE=', 'CODE_FUNC=policy.DEBUG_ALWAYS', -- no meaningful locations
+		'[%-6s]%s', LOG_GRP_REQDBG_TAG, msg) -- msg should end with newline already
 end)
 ffi.gc(debug_logline_cb, free_cb)
 
+-- LOG_DEBUG without log_trace and without code locations
+local function log_notrace(req, fmt, ...)
+	ffi.C.kr_log_fmt(ffi.C.LOG_GRP_REQDBG, LOG_DEBUG,
+		'CODE_FILE=policy.lua', 'CODE_LINE=', 'CODE_FUNC=', -- no meaningful locations
+		'%s', string.format( -- convert in lua, as integers in C varargs would pass as double
+			'[%-6s][%-6s][%05u.00] ' .. fmt,
+			LOG_GRP_REQDBG_TAG, LOG_GRP_POLICY_TAG, req.uid, ...)
+	)
+end
+
 local debug_logfinish_cb = ffi.cast('trace_callback_f', function (req)
 	jit.off(true, true) -- JIT for (C -> lua)^2 nesting isn't allowed
-	ffi.C.kr_log_req(req, 0, 0, 'dbg',
-		'following rrsets were marked as interesting:\n' ..
+	log_notrace(req, 'following rrsets were marked as interesting:\n%s\n',
 		req:selected_tostring())
 	if req.answer ~= nil then
-		ffi.C.kr_log_req(req, 0, 0, 'dbg',
-			'answer packet:\n' ..
-			tostring(req.answer))
+		log_notrace(req, 'answer packet:\n%s\n', req.answer)
 	else
-		ffi.C.kr_log_req(req, 0, 0, 'dbg',
-			'answer packet DROPPED\n')
+		log_notrace(req, 'answer packet DROPPED\n')
 	end
 end)
 ffi.gc(debug_logfinish_cb, free_cb)
 
 -- log request packet
 function policy.REQTRACE(_, req)
-	ffi.C.kr_log_req(req, 0, 0, 'dbg', 'request packet:\n%s',
-		tostring(req.qsource.packet))
+	log_notrace(req, 'request packet:\n%s', req.qsource.packet)
+end
+
+-- log how the request arrived, notably the client's IP
+function policy.IPTRACE(_, req)
+	if req.qsource.addr == nil then
+		log_notrace(req, 'request packet arrived internally\n')
+	else
+		-- stringify transport flags: struct kr_request_qsource_flags
+		local qf = req.qsource.flags
+		local qf_str = qf.tcp and 'TCP' or 'UDP'
+		if qf.tls  then qf_str = qf_str .. ' + TLS'  end
+		if qf.http then qf_str = qf_str .. ' + HTTP' end
+		if qf.xdp  then qf_str = qf_str .. ' + XDP'  end
+
+		log_notrace(req, 'request packet arrived from %s to %s (%s)\n',
+			req.qsource.addr, req.qsource.dst_addr, qf_str)
+	end
+	return nil -- chain rule
 end
 
 function policy.DEBUG_ALWAYS(state, req)
@@ -686,14 +781,19 @@ local debug_stashlog_cb = ffi.cast('trace_log_f', function (req, msg)
 end)
 ffi.gc(debug_stashlog_cb, free_cb)
 
--- buffer verbose logs and print then only if test() returns a truthy value
+-- buffer debug logs and print then only if test() returns a truthy value
 function policy.DEBUG_IF(test)
 	local debug_finish_cb = ffi.cast('trace_callback_f', function (cbreq)
 		jit.off(true, true) -- JIT for (C -> lua)^2 nesting isn't allowed
 		if test(cbreq) then
 			debug_logfinish_cb(cbreq)  -- unconditional version
+
 			local stash = cbreq:vars()['policy_debug_stash']
-			io.write(table.concat(stash, ''))
+			for _, line in ipairs(stash) do -- don't want one huge entry
+				ffi.C.kr_log_fmt(ffi.C.LOG_GRP_REQDBG, LOG_DEBUG,
+					'CODE_FILE=policy.lua', 'CODE_LINE=', 'CODE_FUNC=', -- no meaningful locations
+					'[%-6s]%s', LOG_GRP_REQDBG_TAG, line)
+			end
 		end
 	end)
 	ffi.gc(debug_finish_cb, function (func) func:free() end)
@@ -718,6 +818,14 @@ policy.DENY = policy.DENY_MSG() -- compatibility with < 2.0
 function policy.DROP(_, req)
 	local answer = answer_clear(req)
 	if answer == nil then return nil end
+	req:set_extended_error(kres.extended_error.PROHIBITED, "U5KL")
+	log_policy_action(req, 'DROP')
+	return kres.FAIL
+end
+
+function policy.NO_ANSWER(_, req)
+	req.options.NO_ANSWER = true
+	log_policy_action(req, 'NO_ANSWER')
 	return kres.FAIL
 end
 
@@ -726,6 +834,8 @@ function policy.REFUSE(_, req)
 	if answer == nil then return nil end
 	answer:rcode(kres.rcode.REFUSED)
 	answer:ad(false)
+	req:set_extended_error(kres.extended_error.PROHIBITED, "EIM4")
+	log_policy_action(req, 'REFUSE')
 	return kres.DONE
 end
 
@@ -739,6 +849,7 @@ function policy.TC(state, req)
 	if answer == nil then return nil end
 	answer:tc(1)
 	answer:ad(false)
+	log_policy_action(req, 'TC')
 	return kres.DONE
 end
 
@@ -928,7 +1039,8 @@ policy.special_names = {
 		cb=policy.suffix_common(policy.DENY_MSG(
 			'Blocking is mandated by standards, see references on '
 			.. 'https://www.iana.org/assignments/'
-			.. 'locally-served-dns-zones/locally-served-dns-zones.xhtml'),
+			.. 'locally-served-dns-zones/locally-served-dns-zones.xhtml',
+			kres.extended_error.NOTSUP),
 			private_zones, todname('arpa.')),
 		count=0
 	},
@@ -936,7 +1048,8 @@ policy.special_names = {
 		cb=policy.suffix(policy.DENY_MSG(
 			'Blocking is mandated by standards, see references on '
 			.. 'https://www.iana.org/assignments/'
-			.. 'special-use-domain-names/special-use-domain-names.xhtml'),
+			.. 'special-use-domain-names/special-use-domain-names.xhtml',
+			kres.extended_error.NOTSUP),
 			{
 				todname('test.'),
 				todname('onion.'),
